@@ -2,13 +2,18 @@
 
 set -e
 
+if [ ! -e grub-core/lib/gnulib/stdlib.in.h ]; then
+  echo "Gnulib not yet bootstrapped; run ./bootstrap instead." >&2
+  exit 1
+fi
+
 # Set ${PYTHON} to plain 'python' if not set already
 : ${PYTHON:=python}
 
 export LC_COLLATE=C
 unset LC_ALL
 
-find . -iname '*.[ch]' ! -ipath './grub-core/lib/libgcrypt-grub/*' ! -ipath './build-aux/*' ! -ipath './grub-core/lib/libgcrypt/src/misc.c' ! -ipath './grub-core/lib/libgcrypt/src/global.c' ! -ipath './grub-core/lib/libgcrypt/src/secmem.c'  ! -ipath './util/grub-gen-widthspec.c' ! -ipath './util/grub-gen-asciih.c' |sort > po/POTFILES.in
+find . -iname '*.[ch]' ! -ipath './grub-core/lib/libgcrypt-grub/*' ! -ipath './build-aux/*' ! -ipath './grub-core/lib/libgcrypt/src/misc.c' ! -ipath './grub-core/lib/libgcrypt/src/global.c' ! -ipath './grub-core/lib/libgcrypt/src/secmem.c'  ! -ipath './util/grub-gen-widthspec.c' ! -ipath './util/grub-gen-asciih.c' ! -ipath './gnulib/*' ! -ipath './grub-core/lib/gnulib/*' |sort > po/POTFILES.in
 find util -iname '*.in' ! -name Makefile.in  |sort > po/POTFILES-shell.in
 
 echo "Importing unicode..."
@@ -82,6 +87,17 @@ done
 echo "Saving timestamps..."
 echo timestamp > stamp-h.in
 
-echo "Running autoreconf..."
-autoreconf -vi
+if [ -z "$FROM_BOOTSTRAP" ]; then
+  # Unaided autoreconf is likely to install older versions of many files
+  # than the ones provided by Gnulib, but in most cases this won't matter
+  # very much.  This mode is provided so that you can run ./autogen.sh to
+  # regenerate the GRUB build system in an unpacked release tarball (perhaps
+  # after patching it), even on systems that don't have access to
+  # gnulib.git.
+  echo "Running autoreconf..."
+  cp -a INSTALL INSTALL.grub
+  autoreconf -vif
+  mv INSTALL.grub INSTALL
+fi
+
 exit 0
