@@ -88,7 +88,7 @@ struct grub_nv_super
   char prodrev[NV_PRODREV_LEN];	/* 0x2C - 0x2F Array product revision */
   grub_uint32_t unit_flags;	/* 0x30 - 0x33 Flags for this disk */
   struct grub_nv_array array;	/* Array information */
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 static struct grub_diskfilter_vg *
 grub_dmraid_nv_detect (grub_disk_t disk,
@@ -98,15 +98,16 @@ grub_dmraid_nv_detect (grub_disk_t disk,
   grub_disk_addr_t sector;
   struct grub_nv_super sb;
   int level;
-  int layout;
   grub_uint64_t disk_size;
+  grub_uint32_t capacity;
+  grub_uint8_t total_volumes;
   char *uuid;
 
   if (disk->partition)
     /* Skip partition.  */
     return NULL;
 
-  sector = grub_disk_get_size (disk);
+  sector = grub_disk_native_sectors (disk);
   if (sector == GRUB_DISK_SIZE_UNKNOWN)
     /* Not raid.  */
     return NULL;
@@ -121,15 +122,21 @@ grub_dmraid_nv_detect (grub_disk_t disk,
   if (sb.version != NV_VERSION)
     {
       grub_error (GRUB_ERR_NOT_IMPLEMENTED_YET,
-		  "unknown version: %d.%d", sb.version);
+		  "unknown version: %d.%d", sb.version >> 8, sb.version & 0xFF);
       return NULL;
     }
 
+  capacity = grub_le_to_cpu32 (sb.capacity);
+  total_volumes = sb.array.total_volumes;
+
   switch (sb.array.raid_level)
     {
     case NV_LEVEL_0:
       level = 0;
-      disk_size = sb.capacity / sb.array.total_volumes;
+      if (total_volumes == 0)
+	/* Not RAID.  */
+	return NULL;
+      disk_size = capacity / total_volumes;
       break;
 
     case NV_LEVEL_1:
@@ -139,8 +146,10 @@ grub_dmraid_nv_detect (grub_disk_t disk,
 
     case NV_LEVEL_5:
       level = 5;
-      layout = GRUB_RAID_LAYOUT_LEFT_ASYMMETRIC;
-      disk_size = sb.capacity / (sb.array.total_volumes - 1);
+      if (total_volumes == 0 || total_volumes == 1)
+	/* Not RAID.  */
+	return NULL;
+      disk_size = capacity / (total_volumes - 1);
       break;
 
     default:
@@ -163,8 +172,9 @@ grub_dmraid_nv_detect (grub_disk_t disk,
 
   return grub_diskfilter_make_raid (sizeof (sb.array.signature),
 				    uuid, sb.array.total_volumes,
-				    NULL, disk_size,
-				    sb.array.stripe_block_size, layout,
+				    "nv", disk_size,
+				    sb.array.stripe_block_size,
+				    GRUB_RAID_LAYOUT_LEFT_ASYMMETRIC,
 				    level);
 }
 
