@@ -32,9 +32,11 @@ struct grub_loopback
   char *devname;
   grub_file_t file;
   struct grub_loopback *next;
+  unsigned long id;
 };
 
 static struct grub_loopback *loopback_list;
+static unsigned long last_id = 0;
 
 static const struct grub_arg_option options[] =
   {
@@ -90,22 +92,15 @@ grub_cmd_loopback (grub_extcmd_context_t ctxt, int argc, char **args)
   if (argc < 2)
     return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("filename expected"));
 
-  file = grub_file_open (args[1]);
-  if (! file)
-    return grub_errno;
-
-  /* First try to replace the old device.  */
+  /* Check that a device with requested name does not already exist. */
   for (newdev = loopback_list; newdev; newdev = newdev->next)
     if (grub_strcmp (newdev->devname, args[0]) == 0)
-      break;
+      return grub_error (GRUB_ERR_BAD_ARGUMENT, "device name already exists");
 
-  if (newdev)
-    {
-      grub_file_close (newdev->file);
-      newdev->file = file;
-
-      return 0;
-    }
+  file = grub_file_open (args[1], GRUB_FILE_TYPE_LOOPBACK
+			 | GRUB_FILE_TYPE_NO_DECOMPRESS);
+  if (! file)
+    return grub_errno;
 
   /* Unable to replace it, make a new entry.  */
   newdev = grub_malloc (sizeof (struct grub_loopback));
@@ -120,6 +115,7 @@ grub_cmd_loopback (grub_extcmd_context_t ctxt, int argc, char **args)
     }
 
   newdev->file = file;
+  newdev->id = last_id++;
 
   /* Add the new entry to the list.  */
   newdev->next = loopback_list;
@@ -135,15 +131,15 @@ fail:
 
 
 static int
-grub_loopback_iterate (int (*hook) (const char *name),
-		    grub_disk_pull_t pull)
+grub_loopback_iterate (grub_disk_dev_iterate_hook_t hook, void *hook_data,
+		       grub_disk_pull_t pull)
 {
   struct grub_loopback *d;
   if (pull != GRUB_DISK_PULL_NONE)
     return 0;
   for (d = loopback_list; d; d = d->next)
     {
-      if (hook (d->devname))
+      if (hook (d->devname, hook_data))
 	return 1;
     }
   return 0;
@@ -167,7 +163,11 @@ grub_loopback_open (const char *name, grub_disk_t disk)
 			   / GRUB_DISK_SECTOR_SIZE);
   else
     disk->total_sectors = GRUB_DISK_SIZE_UNKNOWN;
-  disk->id = (unsigned long) dev;
+  /* Avoid reading more than 512M.  */
+  disk->max_agglomerate = 1 << (29 - GRUB_DISK_SECTOR_BITS
+				- GRUB_DISK_CACHE_BITS);
+
+  disk->id = dev->id;
 
   disk->data = dev;
 
@@ -206,17 +206,18 @@ grub_loopback_write (grub_disk_t disk __attribute ((unused)),
 		     grub_size_t size __attribute ((unused)),
 		     const char *buf __attribute ((unused)))
 {
-  return GRUB_ERR_NOT_IMPLEMENTED_YET;
+  return grub_error (GRUB_ERR_NOT_IMPLEMENTED_YET,
+		     "loopback write is not supported");
 }
 
 static struct grub_disk_dev grub_loopback_dev =
   {
     .name = "loopback",
     .id = GRUB_DISK_DEVICE_LOOPBACK_ID,
-    .iterate = grub_loopback_iterate,
-    .open = grub_loopback_open,
-    .read = grub_loopback_read,
-    .write = grub_loopback_write,
+    .disk_iterate = grub_loopback_iterate,
+    .disk_open = grub_loopback_open,
+    .disk_read = grub_loopback_read,
+    .disk_write = grub_loopback_write,
     .next = 0
   };
 
