@@ -44,7 +44,7 @@ struct grub_usbms_cbw
   grub_uint8_t lun;
   grub_uint8_t length;
   grub_uint8_t cbwcb[16];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_usbms_csw
 {
@@ -52,7 +52,7 @@ struct grub_usbms_csw
   grub_uint32_t tag;
   grub_uint32_t residue;
   grub_uint8_t status;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_usbms_dev
 {
@@ -76,7 +76,7 @@ typedef struct grub_usbms_dev *grub_usbms_dev_t;
 static grub_usbms_dev_t grub_usbms_devices[MAX_USBMS_DEVICES];
 static int first_available_slot = 0;
 
-static grub_err_t
+static grub_usb_err_t
 grub_usbms_cbi_cmd (grub_usb_device_t dev, int interface,
                     grub_uint8_t *cbicb)
 {
@@ -86,7 +86,7 @@ grub_usbms_cbi_cmd (grub_usb_device_t dev, int interface,
                                GRUB_USBMS_CBI_CMD_SIZE, (char*)cbicb);
 }
 
-static grub_err_t
+static grub_usb_err_t
 grub_usbms_cbi_reset (grub_usb_device_t dev, int interface)
 {
   /* Prepare array with Command Block Reset (=CBR) */
@@ -108,17 +108,13 @@ grub_usbms_cbi_reset (grub_usb_device_t dev, int interface)
   return grub_usbms_cbi_cmd (dev, interface, (grub_uint8_t *)&cbicb);
 }
 
-static grub_err_t
+static grub_usb_err_t
 grub_usbms_bo_reset (grub_usb_device_t dev, int interface)
 {
-  grub_usb_err_t u;
-  u = grub_usb_control_msg (dev, 0x21, 255, 0, interface, 0, 0);
-  if (u)
-    return grub_error (GRUB_ERR_IO, "USB error %d", u);
-  return GRUB_ERR_NONE;
+  return grub_usb_control_msg (dev, 0x21, 255, 0, interface, 0, 0);
 }
 
-static grub_err_t
+static grub_usb_err_t
 grub_usbms_reset (grub_usbms_dev_t dev)
 {
   if (dev->protocol == GRUB_USBMS_PROTOCOL_BULK)
@@ -149,7 +145,7 @@ grub_usbms_attach (grub_usb_device_t usbdev, int configno, int interfno)
   int j;
   grub_uint8_t luns = 0;
   unsigned curnum;
-  grub_usb_err_t err = GRUB_ERR_NONE;
+  grub_usb_err_t err = GRUB_USB_ERR_NONE;
 
   grub_boot_time ("Attaching USB mass storage");
 
@@ -302,6 +298,8 @@ grub_usbms_transfer_bo (struct grub_scsi *scsi, grub_size_t cmdsize, char *cmd,
   grub_usb_err_t errCSW = GRUB_USB_ERR_NONE;
   int retrycnt = 3 + 1;
   
+  tag++;
+
  retry:
   retrycnt--;
   if (retrycnt == 0)
@@ -309,8 +307,8 @@ grub_usbms_transfer_bo (struct grub_scsi *scsi, grub_size_t cmdsize, char *cmd,
 
   /* Setup the request.  */
   grub_memset (&cbw, 0, sizeof (cbw));
-  cbw.signature = grub_cpu_to_le32 (0x43425355);
-  cbw.tag = tag++;
+  cbw.signature = grub_cpu_to_le32_compile_time (0x43425355);
+  cbw.tag = tag;
   cbw.transfer_length = grub_cpu_to_le32 (size);
   cbw.flags = (!read_write) << GRUB_USBMS_DIRECTION_BIT;
   cbw.lun = scsi->lun; /* In USB MS CBW are LUN bits on another place than in SCSI CDB, both should be set correctly. */
@@ -339,7 +337,7 @@ grub_usbms_transfer_bo (struct grub_scsi *scsi, grub_size_t cmdsize, char *cmd,
 	  grub_usb_clear_halt (dev->dev, dev->out->endp_addr);
 	  goto CheckCSW;
 	}
-      return grub_error (GRUB_ERR_IO, "USB Mass Storage request failed");
+      goto retry;
     }
 
   /* Read/write the data, (maybe) according to specification.  */
@@ -416,7 +414,7 @@ CheckCSW:
   
   /* If phase error or not valid signature, do bulk-only reset device. */
   if ((status.status == 2) ||
-      (status.signature != grub_cpu_to_le32(0x53425355)))
+      (status.signature != grub_cpu_to_le32_compile_time(0x53425355)))
     { /* Bulk-only reset device. */
       grub_dprintf ("usb", "Bulk-only reset device - bad status\n");
       grub_usbms_reset (dev);
@@ -571,8 +569,11 @@ grub_usbms_transfer_cbi (struct grub_scsi *scsi, grub_size_t cmdsize, char *cmd,
             }
         }
     }
+
+  if (err)
+    return grub_error (GRUB_ERR_IO, "USB error %d", err);
     
-  return err;
+  return GRUB_ERR_NONE;
 }
 
 
