@@ -36,7 +36,7 @@ struct grub_affs_bblock
   grub_uint8_t flags;
   grub_uint32_t checksum;
   grub_uint32_t rootblock;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* Set if the filesystem is a AFFS filesystem.  Otherwise this is an
    OFS filesystem.  */
@@ -51,14 +51,14 @@ struct grub_affs_rblock
   grub_uint32_t unused2;
   grub_uint32_t checksum;
   grub_uint32_t hashtable[1];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_affs_time
 {
   grub_int32_t day;
   grub_uint32_t min;
   grub_uint32_t hz;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* The second part of a file header block.  */
 struct grub_affs_file
@@ -76,7 +76,7 @@ struct grub_affs_file
   grub_uint32_t parent;
   grub_uint32_t extension;
   grub_uint32_t type;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* The location of `struct grub_affs_file' relative to the end of a
    file header block.  */
@@ -301,7 +301,7 @@ grub_affs_read_symlink (grub_fshelp_node_t node)
       return 0;
     }
   latin1[symlink_size] = 0;
-  utf8 = grub_malloc (symlink_size * GRUB_MAX_UTF8_PER_LATIN1 + 1);
+  utf8 = grub_calloc (GRUB_MAX_UTF8_PER_LATIN1 + 1, symlink_size);
   if (!utf8)
     {
       grub_free (latin1);
@@ -325,7 +325,7 @@ grub_affs_create_node (grub_fshelp_node_t dir,
 		       grub_uint32_t block, const struct grub_affs_file *fil)
 {
   struct grub_affs_data *data = dir->data;
-  int type;
+  int type = GRUB_FSHELP_REG;
   grub_uint8_t name_u8[sizeof (fil->name) * GRUB_MAX_UTF8_PER_LATIN1 + 1];
   grub_size_t len;
   unsigned int nest;
@@ -400,12 +400,12 @@ grub_affs_iterate_dir (grub_fshelp_node_t dir,
 {
   unsigned int i;
   struct grub_affs_file file;
-  struct grub_fshelp_node *node = 0;
+  struct grub_fshelp_node *node, *orig_node;
   struct grub_affs_data *data = dir->data;
   grub_uint32_t *hashtable;
 
   /* Create the directory entries for `.' and `..'.  */
-  node = grub_zalloc (sizeof (*node));
+  node = orig_node = grub_zalloc (sizeof (*node));
   if (!node)
     return 1;
     
@@ -414,15 +414,12 @@ grub_affs_iterate_dir (grub_fshelp_node_t dir,
     return 1;
   if (dir->parent)
     {
-      node = grub_zalloc (sizeof (*node));
-      if (!node)
-	return 1;
       *node = *dir->parent;
       if (hook ("..", GRUB_FSHELP_DIR, node, hook_data))
 	return 1;
     }
 
-  hashtable = grub_zalloc (data->htsize * sizeof (*hashtable));
+  hashtable = grub_calloc (data->htsize, sizeof (*hashtable));
   if (!hashtable)
     return 1;
 
@@ -456,17 +453,18 @@ grub_affs_iterate_dir (grub_fshelp_node_t dir,
 
 	  if (grub_affs_create_node (dir, hook, hook_data, &node, &hashtable,
 				     next, &file))
-	    return 1;
+	    {
+	      /* Node has been replaced in function. */
+	      grub_free (orig_node);
+	      return 1;
+	    }
 
 	  next = grub_be_to_cpu32 (file.next);
 	}
     }
 
-  grub_free (hashtable);
-  return 0;
-
  fail:
-  grub_free (node);
+  grub_free (orig_node);
   grub_free (hashtable);
   return 0;
 }
@@ -628,7 +626,7 @@ grub_affs_label (grub_device_t device, char **label)
       len = file.namelen;
       if (len > sizeof (file.name))
 	len = sizeof (file.name);
-      *label = grub_malloc (len * GRUB_MAX_UTF8_PER_LATIN1 + 1);
+      *label = grub_calloc (GRUB_MAX_UTF8_PER_LATIN1 + 1, len);
       if (*label)
 	*grub_latin1_to_utf8 ((grub_uint8_t *) *label, file.name, len) = '\0';
     }
@@ -643,7 +641,7 @@ grub_affs_label (grub_device_t device, char **label)
 }
 
 static grub_err_t
-grub_affs_mtime (grub_device_t device, grub_int32_t *t)
+grub_affs_mtime (grub_device_t device, grub_int64_t *t)
 {
   struct grub_affs_data *data;
   grub_disk_t disk = device->disk;
@@ -685,12 +683,12 @@ grub_affs_mtime (grub_device_t device, grub_int32_t *t)
 static struct grub_fs grub_affs_fs =
   {
     .name = "affs",
-    .dir = grub_affs_dir,
-    .open = grub_affs_open,
-    .read = grub_affs_read,
-    .close = grub_affs_close,
-    .label = grub_affs_label,
-    .mtime = grub_affs_mtime,
+    .fs_dir = grub_affs_dir,
+    .fs_open = grub_affs_open,
+    .fs_read = grub_affs_read,
+    .fs_close = grub_affs_close,
+    .fs_label = grub_affs_label,
+    .fs_mtime = grub_affs_mtime,
 
 #ifdef GRUB_UTIL
     .reserved_first_sector = 0,
