@@ -40,6 +40,8 @@
 
 static void
 init_terminal (grub_gfxmenu_view_t view);
+static void
+init_background (grub_gfxmenu_view_t view);
 static grub_gfxmenu_view_t term_view;
 
 /* Create a new view object, loading the theme specified by THEME_PATH and
@@ -91,7 +93,11 @@ grub_gfxmenu_view_new (const char *theme_path,
   view->title_color = default_fg_color;
   view->message_color = default_bg_color;
   view->message_bg_color = default_fg_color;
-  view->desktop_image = 0;
+  view->raw_desktop_image = 0;
+  view->scaled_desktop_image = 0;
+  view->desktop_image_scale_method = GRUB_VIDEO_BITMAP_SELECTION_METHOD_STRETCH;
+  view->desktop_image_h_align = GRUB_VIDEO_BITMAP_H_ALIGN_CENTER;
+  view->desktop_image_v_align = GRUB_VIDEO_BITMAP_V_ALIGN_CENTER;
   view->desktop_color = default_bg_color;
   view->terminal_box = grub_gfxmenu_create_box (0, 0);
   view->title_text = grub_strdup (_("GRUB Boot Menu"));
@@ -128,7 +134,8 @@ grub_gfxmenu_view_destroy (grub_gfxmenu_view_t view)
       grub_gfxmenu_timeout_notifications = grub_gfxmenu_timeout_notifications->next;
       grub_free (p);
     }
-  grub_video_bitmap_destroy (view->desktop_image);
+  grub_video_bitmap_destroy (view->raw_desktop_image);
+  grub_video_bitmap_destroy (view->scaled_desktop_image);
   if (view->terminal_box)
     view->terminal_box->destroy (view->terminal_box);
   grub_free (view->terminal_font_name);
@@ -144,9 +151,9 @@ static void
 redraw_background (grub_gfxmenu_view_t view,
 		   const grub_video_rect_t *bounds)
 {
-  if (view->desktop_image)
+  if (view->scaled_desktop_image)
     {
-      struct grub_video_bitmap *img = view->desktop_image;
+      struct grub_video_bitmap *img = view->scaled_desktop_image;
       grub_video_blit_bitmap (img, GRUB_VIDEO_BLIT_REPLACE,
                               bounds->x, bounds->y,
 			      bounds->x - view->screen.x,
@@ -206,6 +213,7 @@ redraw_timeouts (struct grub_gfxmenu_view *view)
     {
       grub_video_rect_t bounds;
       cur->self->ops->get_bounds (cur->self, &bounds);
+      grub_video_set_area_status (GRUB_VIDEO_AREA_ENABLED);
       grub_gfxmenu_view_redraw (view, &bounds);
     }
 }
@@ -314,6 +322,11 @@ grub_gfxmenu_view_redraw (grub_gfxmenu_view_t view,
     grub_gfxterm_schedule_repaint ();
 
   grub_video_set_active_render_target (GRUB_VIDEO_RENDER_TARGET_DISPLAY);
+  grub_video_area_status_t area_status;
+  grub_video_get_area_status (&area_status);
+  if (area_status == GRUB_VIDEO_AREA_ENABLED)
+    grub_video_set_region (region->x, region->y,
+                           region->width, region->height);
 
   redraw_background (view, region);
   if (view->canvas)
@@ -321,6 +334,9 @@ grub_gfxmenu_view_redraw (grub_gfxmenu_view_t view,
   draw_title (view);
   if (grub_video_have_common_points (&view->progress_message_frame, region))
     draw_message (view);
+
+  if (area_status == GRUB_VIDEO_AREA_ENABLED)
+    grub_video_set_area_status (GRUB_VIDEO_AREA_ENABLED);
 }
 
 void
@@ -328,6 +344,8 @@ grub_gfxmenu_view_draw (grub_gfxmenu_view_t view)
 {
   init_terminal (view);
 
+  init_background (view);
+
   /* Clear the screen; there may be garbage left over in video memory. */
   grub_video_fill_rect (grub_video_map_rgb (0, 0, 0),
                         view->screen.x, view->screen.y,
@@ -341,10 +359,15 @@ grub_gfxmenu_view_draw (grub_gfxmenu_view_t view)
   refresh_menu_components (view);
   update_menu_components (view);
 
+  grub_video_set_area_status (GRUB_VIDEO_AREA_DISABLED);
   grub_gfxmenu_view_redraw (view, &view->screen);
   grub_video_swap_buffers ();
   if (view->double_repaint)
-    grub_gfxmenu_view_redraw (view, &view->screen);
+    {
+      grub_video_set_area_status (GRUB_VIDEO_AREA_DISABLED);
+      grub_gfxmenu_view_redraw (view, &view->screen);
+    }
+
 }
 
 static void
@@ -358,6 +381,7 @@ redraw_menu_visit (grub_gui_component_t component,
       grub_video_rect_t bounds;
 
       component->ops->get_bounds (component, &bounds);
+      grub_video_set_area_status (GRUB_VIDEO_AREA_ENABLED);
       grub_gfxmenu_view_redraw (view, &bounds);
     }
 }
@@ -395,6 +419,8 @@ grub_gfxmenu_draw_terminal_box (void)
   if (!term_box)
     return;
 
+  grub_video_set_area_status (GRUB_VIDEO_AREA_DISABLED);
+
   term_box->set_content_size (term_box, term_view->terminal_rect.width,
 			      term_view->terminal_rect.height);
   
@@ -524,6 +550,35 @@ init_terminal (grub_gfxmenu_view_t view)
   grub_gfxterm_decorator_hook = grub_gfxmenu_draw_terminal_box;
 }
 
+static void
+init_background (grub_gfxmenu_view_t view)
+{
+  if (view->scaled_desktop_image)
+    return;
+
+  struct grub_video_bitmap *scaled_bitmap;
+  if (view->desktop_image_scale_method ==
+      GRUB_VIDEO_BITMAP_SELECTION_METHOD_STRETCH)
+    grub_video_bitmap_create_scaled (&scaled_bitmap,
+                                     view->screen.width,
+                                     view->screen.height,
+                                     view->raw_desktop_image,
+                                     GRUB_VIDEO_BITMAP_SCALE_METHOD_BEST);
+  else
+    grub_video_bitmap_scale_proportional (&scaled_bitmap,
+                                          view->screen.width,
+                                          view->screen.height,
+                                          view->raw_desktop_image,
+                                          GRUB_VIDEO_BITMAP_SCALE_METHOD_BEST,
+                                          view->desktop_image_scale_method,
+                                          view->desktop_image_v_align,
+                                          view->desktop_image_h_align);
+  if (! scaled_bitmap)
+    return;
+  view->scaled_desktop_image = scaled_bitmap;
+
+}
+
 /* FIXME: previously notifications were displayed in special case.
    Is it necessary?
  */
