@@ -85,9 +85,9 @@ heap_init (grub_uint64_t addr, grub_uint64_t size, grub_memory_type_t type,
 struct resource
 {
   grub_pci_device_t dev;
-  int type;
   grub_size_t size;
-  int bar;
+  unsigned type:4;
+  unsigned bar:3;
 };
 
 struct iterator_ctx
@@ -96,17 +96,8 @@ struct iterator_ctx
   grub_size_t nresources;
 };
 
-static int
-count_cards (grub_pci_device_t dev __attribute__ ((unused)),
-	     grub_pci_id_t pciid __attribute__ ((unused)),
-	     void *data)
-{
-  int *cnt = data;
-
-  (*cnt)++;
-
-  return 0;
-}
+/* We don't support bridges, so can't have more than 32 devices.  */
+#define MAX_DEVICES 32
 
 static int
 find_resources (grub_pci_device_t dev,
@@ -116,6 +107,9 @@ find_resources (grub_pci_device_t dev,
   struct iterator_ctx *ctx = data;
   int bar;
 
+  if (ctx->nresources >= MAX_DEVICES * 6)
+    return 1;
+
   for (bar = 0; bar < 6; bar++)
     {
       grub_pci_address_t addr;
@@ -190,13 +184,10 @@ enable_cards (grub_pci_device_t dev,
 static void
 grub_pci_assign_addresses (void)
 {
-  int ncards = 0;
   struct iterator_ctx ctx;
 
-  grub_pci_iterate (count_cards, &ncards);
-
   {
-    struct resource resources[ncards * 6];
+    struct resource resources[MAX_DEVICES * 6];
     int done;
     unsigned i;
     ctx.nresources = 0;
@@ -277,8 +268,9 @@ grub_machine_get_bootlocation (char **device __attribute__ ((unused)),
 }
 
 void
-grub_machine_fini (void)
+grub_machine_fini (int flags)
 {
-  grub_vga_text_fini ();
+  if (flags & GRUB_LOADER_FLAG_NORETURN)
+    grub_vga_text_fini ();
   grub_stop_floppy ();
 }
