@@ -19,11 +19,12 @@
 #include <grub/kernel.h>
 #include <grub/misc.h>
 #include <grub/uboot/uboot.h>
+#include <grub/loader.h>
 
 void
 grub_reboot (void)
 {
-  grub_machine_fini ();
+  grub_machine_fini (GRUB_LOADER_FLAG_NORETURN);
 
   grub_uboot_reset ();
   while (1);
