@@ -24,6 +24,7 @@
 #include <grub/mm.h>
 #include <grub/script_sh.h>
 #include <grub/i18n.h>
+#include <grub/safemath.h>
 
 #define yytext_ptr char *
 #include "grub_script.tab.h"
@@ -110,10 +111,14 @@ grub_script_lexer_record (struct grub_parser_param *parser, char *str)
       old = lexer->recording;
       if (lexer->recordlen < len)
 	lexer->recordlen = len;
-      lexer->recordlen *= 2;
+
+      if (grub_mul (lexer->recordlen, 2, &lexer->recordlen))
+	goto fail;
+
       lexer->recording = grub_realloc (lexer->recording, lexer->recordlen);
       if (!lexer->recording)
 	{
+ fail:
 	  grub_free (old);
 	  lexer->recordpos = 0;
 	  lexer->recordlen = 0;
@@ -130,7 +135,7 @@ int
 grub_script_lexer_yywrap (struct grub_parser_param *parserstate,
 			  const char *input)
 {
-  grub_size_t len = 0;
+  grub_size_t len = 0, sz;
   char *p = 0;
   char *line = 0;
   YY_BUFFER_STATE buffer;
@@ -147,7 +152,7 @@ grub_script_lexer_yywrap (struct grub_parser_param *parserstate,
 
   line = 0;
   if (! input)
-    lexerstate->getline (&line, 1);
+    lexerstate->getline (&line, 1, lexerstate->getline_data);
   else
     line = grub_strdup (input);
 
@@ -164,15 +169,26 @@ grub_script_lexer_yywrap (struct grub_parser_param *parserstate,
     {
       grub_free (line);
       line = grub_strdup ("\n");
+      len = 1;
     }
   else if (len && line[len - 1] != '\n')
     {
-      p = grub_realloc (line, len + 2);
+      if (grub_add (len, 2, &sz))
+	{
+	  grub_free (line);
+	  grub_script_yyerror (parserstate, N_("overflow is detected"));
+	  return 1;
+	}
+
+      p = grub_realloc (line, sz);
       if (p)
 	{
 	  p[len++] = '\n';
 	  p[len] = '\0';
 	}
+      else
+	grub_free (line);
+
       line = p;
     }
 
@@ -216,7 +232,7 @@ grub_script_lexer_yywrap (struct grub_parser_param *parserstate,
 
 struct grub_lexer_param *
 grub_script_lexer_init (struct grub_parser_param *parser, char *script,
-			grub_reader_getline_t arg_getline)
+			grub_reader_getline_t arg_getline, void *getline_data)
 {
   struct grub_lexer_param *lexerstate;
 
@@ -232,7 +248,10 @@ grub_script_lexer_init (struct grub_parser_param *parser, char *script,
       return 0;
     }
 
-  lexerstate->getline = arg_getline;	/* rest are all zeros already */
+  lexerstate->getline = arg_getline;
+  lexerstate->getline_data = getline_data;
+  /* The other elements of lexerstate are all zeros already.  */
+
   if (yylex_init (&lexerstate->yyscanner))
     {
       grub_free (lexerstate->text);
@@ -247,7 +266,6 @@ grub_script_lexer_init (struct grub_parser_param *parser, char *script,
     {
       parser->lexerstate = 0;
       yylex_destroy (lexerstate->yyscanner);
-      grub_free (lexerstate->yyscanner);
       grub_free (lexerstate->text);
       grub_free (lexerstate);
       return 0;
@@ -328,10 +346,10 @@ grub_script_yylex (union YYSTYPE *value,
 }
 
 void
-grub_script_yyerror (struct grub_parser_param *state, char const *err)
+grub_script_yyerror (struct grub_parser_param *state, const char *err)
 {
   if (err)
-    grub_error (GRUB_ERR_INVALID_COMMAND, err);
+    grub_error (GRUB_ERR_INVALID_COMMAND, "%s", err);
 
   grub_print_error ();
   state->err++;
