@@ -17,12 +17,19 @@
  */
 
 /* All tests need to include test.h for GRUB testing framework.  */
+
+#include <config.h>
+
 #include <grub/test.h>
 #include <grub/dl.h>
 #include <grub/video.h>
 #include <grub/lib/crc.h>
 #include <grub/mm.h>
 #include <grub/term.h>
+#ifdef GRUB_MACHINE_EMU
+#include <grub/emu/hostdisk.h>
+#include <grub/emu/misc.h>
+#endif
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
@@ -32,7 +39,7 @@ static char *basename;
 static const grub_uint32_t *checksums;
 static struct grub_video_mode_info capt_mode_info;
 
-struct grub_video_mode_info grub_test_video_modes[30] = {
+struct grub_video_mode_info grub_test_video_modes[GRUB_TEST_VIDEO_ALL_N_MODES] = {
   {
     .width = 640,
     .height = 480,
@@ -78,7 +85,12 @@ struct grub_video_mode_info grub_test_video_modes[30] = {
     .pitch = 1024 * 4,
     GRUB_VIDEO_MI_RGBA8888()
   },
-
+  {
+    .width = 2560,
+    .height = 1440,
+    .pitch = 2560 * 4,
+    GRUB_VIDEO_MI_RGBA8888()
+  },
   {
     .width = 640,
     .height = 480,
@@ -239,14 +251,7 @@ struct grub_video_mode_info grub_test_video_modes[30] = {
 };
 
 #ifdef GRUB_MACHINE_EMU
-#include <grub/emu/misc.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
-#include <unistd.h>
-#include <errno.h>
-#include <string.h>
-#include <stdlib.h>
+#include <grub/emu/hostfile.h>
 
 struct bmp_header
 {
@@ -259,20 +264,20 @@ struct bmp_header
   grub_uint16_t height;
   grub_uint16_t planes;
   grub_uint16_t bpp;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 static void
 grub_video_capture_write_bmp (const char *fname,
 			      void *ptr,
 			      const struct grub_video_mode_info *mode_info)
 {
-  int fd = open (fname, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
+  grub_util_fd_t fd = grub_util_fd_open (fname, GRUB_UTIL_FD_O_WRONLY | GRUB_UTIL_FD_O_CREATTRUNC);
   struct bmp_header head;
 
-  if (fd < 0)
+  if (!GRUB_UTIL_FD_IS_VALID (fd))
     {
       grub_printf (_("cannot open `%s': %s"),
-		   fname, strerror (errno));
+		   fname, grub_util_fd_strerror ());
     }
 
   grub_memset (&head, 0, sizeof (head));
@@ -283,16 +288,16 @@ grub_video_capture_write_bmp (const char *fname,
   if (mode_info->mode_type & GRUB_VIDEO_MODE_TYPE_RGB)
     {
       head.filesize = grub_cpu_to_le32 (sizeof (head) + mode_info->width * mode_info->height * 3);
-      head.bmp_off = grub_cpu_to_le32 (sizeof (head));
+      head.bmp_off = grub_cpu_to_le32_compile_time (sizeof (head));
       head.bpp = grub_cpu_to_le16_compile_time (24);
     }
   else
     {
       head.filesize = grub_cpu_to_le32 (sizeof (head) + 3 * 256 + mode_info->width * mode_info->height);
-      head.bmp_off = grub_cpu_to_le32 (sizeof (head) + 3 * 256);
+      head.bmp_off = grub_cpu_to_le32_compile_time (sizeof (head) + 3 * 256);
       head.bpp = grub_cpu_to_le16_compile_time (8);
     }
-  head.head_size = grub_cpu_to_le32 (sizeof (head) - 14);
+  head.head_size = grub_cpu_to_le32_compile_time (sizeof (head) - 14);
   head.width = grub_cpu_to_le16 (mode_info->width);
   head.height = grub_cpu_to_le16 (mode_info->height);
   head.planes = grub_cpu_to_le16_compile_time (1);
@@ -300,7 +305,7 @@ grub_video_capture_write_bmp (const char *fname,
   head.width = grub_cpu_to_le16 (mode_info->width);
   head.height = grub_cpu_to_le16 (mode_info->height);
 
-  write (fd, &head, sizeof (head));
+  grub_util_fd_write (fd, (char *) &head, sizeof (head));
 
   if (!(mode_info->mode_type & GRUB_VIDEO_MODE_TYPE_RGB))
     {
@@ -318,9 +323,9 @@ grub_video_capture_write_bmp (const char *fname,
 	  g = palette_data[i].g;
 	  b = palette_data[i].b;
 
-	  write (fd, &b, 1);
-	  write (fd, &g, 1);
-	  write (fd, &r, 1);
+	  grub_util_fd_write (fd, (char *) &b, 1);
+	  grub_util_fd_write (fd, (char *) &g, 1);
+	  grub_util_fd_write (fd, (char *) &r, 1);
 	}
     }
 
@@ -331,7 +336,7 @@ grub_video_capture_write_bmp (const char *fname,
     {
     case 4:
       {
-	grub_uint8_t *buffer = xmalloc (mode_info->width * 3);
+	grub_uint8_t *buffer = xcalloc (3, mode_info->width);
 	grub_uint32_t rmask = ((1 << mode_info->red_mask_size) - 1);
 	grub_uint32_t gmask = ((1 << mode_info->green_mask_size) - 1);
 	grub_uint32_t bmask = ((1 << mode_info->blue_mask_size) - 1);
@@ -345,7 +350,7 @@ grub_video_capture_write_bmp (const char *fname,
 
 	for (y = mode_info->height - 1; y >= 0; y--)
 	  {
-	    grub_uint32_t *iptr = (grub_uint32_t *) ((grub_uint8_t *) ptr + mode_info->pitch * y);
+	    grub_uint32_t *iptr = (grub_uint32_t *) ptr + (mode_info->pitch / 4) * y;
 	    int x;
 	    grub_uint8_t *optr = buffer;
 	    for (x = 0; x < (int) mode_info->width; x++)
@@ -355,14 +360,14 @@ grub_video_capture_write_bmp (const char *fname,
 		*optr++ = ((val >> gshift) & gmask) << mulgshift;
 		*optr++ = ((val >> rshift) & rmask) << mulrshift;
 	      }
-	    write (fd, buffer, mode_info->width * 3);
+	    grub_util_fd_write (fd, (char *) buffer, mode_info->width * 3);
 	  }
 	grub_free (buffer);
 	break;
       }
     case 3:
       {
-	grub_uint8_t *buffer = xmalloc (mode_info->width * 3);
+	grub_uint8_t *buffer = xcalloc (3, mode_info->width);
 	grub_uint32_t rmask = ((1 << mode_info->red_mask_size) - 1);
 	grub_uint32_t gmask = ((1 << mode_info->green_mask_size) - 1);
 	grub_uint32_t bmask = ((1 << mode_info->blue_mask_size) - 1);
@@ -395,14 +400,14 @@ grub_video_capture_write_bmp (const char *fname,
 		*optr++ = ((val >> gshift) & gmask) << mulgshift;
 		*optr++ = ((val >> rshift) & rmask) << mulrshift;
 	      }
-	    write (fd, buffer, mode_info->width * 3);
+	    grub_util_fd_write (fd, (char *) buffer, mode_info->width * 3);
 	  }
 	grub_free (buffer);
 	break;
       }
     case 2:
       {
-	grub_uint8_t *buffer = xmalloc (mode_info->width * 3);
+	grub_uint8_t *buffer = xcalloc (3, mode_info->width);
 	grub_uint16_t rmask = ((1 << mode_info->red_mask_size) - 1);
 	grub_uint16_t gmask = ((1 << mode_info->green_mask_size) - 1);
 	grub_uint16_t bmask = ((1 << mode_info->blue_mask_size) - 1);
@@ -416,7 +421,7 @@ grub_video_capture_write_bmp (const char *fname,
 
 	for (y = mode_info->height - 1; y >= 0; y--)
 	  {
-	    grub_uint16_t *iptr = (grub_uint16_t *) ((grub_uint8_t *) ptr + mode_info->pitch * y);
+	    grub_uint16_t *iptr = (grub_uint16_t *) ptr + (mode_info->pitch / 2) * y;
 	    int x;
 	    grub_uint8_t *optr = buffer;
 	    for (x = 0; x < (int) mode_info->width; x++)
@@ -426,7 +431,7 @@ grub_video_capture_write_bmp (const char *fname,
 		*optr++ = ((val >> gshift) & gmask) << mulgshift;
 		*optr++ = ((val >> rshift) & rmask) << mulrshift;
 	      }
-	    write (fd, buffer, mode_info->width * 3);
+	    grub_util_fd_write (fd, (char *) buffer, mode_info->width * 3);
 	  }
 	grub_free (buffer);
 	break;
@@ -436,11 +441,12 @@ grub_video_capture_write_bmp (const char *fname,
 	int y;
 
 	for (y = mode_info->height - 1; y >= 0; y--)
-	  write (fd, ((grub_uint8_t *) ptr + mode_info->pitch * y), mode_info->width);
+	  grub_util_fd_write (fd, ((char *) ptr + mode_info->pitch * y),
+			      mode_info->width);
 	break;
       }
     }
-  close (fd);
+  grub_util_fd_close (fd);
 }
 
 #endif
@@ -474,16 +480,11 @@ grub_video_checksum_get_modename (void)
 //#define COLLECT_TIME_STATISTICS 1
 
 #if defined (GENERATE_MODE) && defined (GRUB_MACHINE_EMU)
-int genfd = -1;
+grub_util_fd_t genfd = GRUB_UTIL_FD_INVALID;
 #endif
 
 #include <grub/time.h>
 
-#if defined (GRUB_MACHINE_EMU) && defined (COLLECT_TIME_STATISTICS)
-#include <sys/times.h>
-#include <unistd.h>
-#endif
-
 static void
 write_time (void)
 {
@@ -491,13 +492,12 @@ write_time (void)
   char buf[60];
   static grub_uint64_t prev;
   grub_uint64_t cur;
-  static int tmrfd = -1;
-  struct tms tm;
-  if (tmrfd < 0)
-    tmrfd = open ("time.txt", O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
+  static grub_util_fd_t tmrfd = GRUB_UTIL_FD_INVALID;
+  if (!GRUB_UTIL_FD_IS_VALID (tmrfd))
+    tmrfd = grub_util_fd_open ("time.txt", GRUB_UTIL_FD_O_WRONLY
+			       | GRUB_UTIL_FD_O_CREATTRUNC);
 
-  times (&tm); 
-  cur = (tm.tms_utime * 1000ULL) / sysconf(_SC_CLK_TCK);
+  cur = grub_util_get_cpu_time_ms ();
   grub_snprintf (buf, sizeof (buf), "%s_%dx%dx%s:%d: %" PRIuGRUB_UINT64_T " ms\n",
 		 basename, 			
 		 capt_mode_info.width,
@@ -505,8 +505,8 @@ write_time (void)
 		 grub_video_checksum_get_modename (), ctr,
 		 cur - prev);
   prev = cur;
-  if (tmrfd >= 0)
-    write (tmrfd, buf, grub_strlen (buf));
+  if (GRUB_UTIL_FD_IS_VALID (tmrfd))
+    grub_util_fd_write (tmrfd, buf, grub_strlen (buf));
 #endif
 }
 
@@ -594,11 +594,11 @@ checksum (void)
 #endif
 
 #if defined (GENERATE_MODE) && defined (GRUB_MACHINE_EMU)
-  if (genfd >= 0)
+  if (GRUB_UTIL_FD_IS_VALID (genfd))
     {
       char buf[20];
       grub_snprintf (buf, sizeof (buf), "0x%x, ", crc);
-      write (genfd, buf, grub_strlen (buf));
+      grub_util_fd_write (genfd, buf, grub_strlen (buf));
     }
 #endif
 
@@ -675,9 +675,10 @@ grub_video_checksum (const char *basename_in)
   grub_video_get_info (&capt_mode_info);
 
 #if defined (GENERATE_MODE) && defined (GRUB_MACHINE_EMU)
-  if (genfd < 0)
-    genfd = open ("checksums.h", O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
-  if (genfd >= 0)
+  if (!GRUB_UTIL_FD_IS_VALID (genfd))
+    genfd = grub_util_fd_open ("checksums.h", GRUB_UTIL_FD_O_WRONLY
+			       | GRUB_UTIL_FD_O_CREATTRUNC);
+  if (GRUB_UTIL_FD_IS_VALID (genfd))
     {
       char buf[400];
 
@@ -700,9 +701,9 @@ grub_video_checksum (const char *basename_in)
 		     capt_mode_info.height,
 		     grub_video_checksum_get_modename ());
 
-      write (genfd, "  { \"", 5);
-      write (genfd, basename_in, grub_strlen (basename_in));
-      write (genfd, buf, grub_strlen (buf));
+      grub_util_fd_write (genfd, "  { \"", 5);
+      grub_util_fd_write (genfd, basename_in, grub_strlen (basename_in));
+      grub_util_fd_write (genfd, buf, grub_strlen (buf));
     }
 #endif
 
@@ -740,11 +741,11 @@ void
 grub_video_checksum_end (void)
 {
 #if defined (GENERATE_MODE) && defined (GRUB_MACHINE_EMU)
-  if (genfd >= 0)
+  if (GRUB_UTIL_FD_IS_VALID (genfd))
     {
       char buf[40];
       grub_snprintf (buf, sizeof (buf), "}, %d },\n", ctr);
-      write (genfd, buf, grub_strlen (buf));
+      grub_util_fd_write (genfd, buf, grub_strlen (buf));
     }
 #endif
   grub_test_assert (ctr == nchk, "Not enough checksums %s_%dx%dx%s: %d vs %d",
@@ -764,12 +765,11 @@ grub_video_checksum_end (void)
 static struct grub_term_output *saved_outputs;
 static struct grub_term_output *saved_gfxnext;
 static struct grub_term_output *gfxterm;
+static int use_gfxterm = 0;
 
 int
 grub_test_use_gfxterm (void)
 {
-  saved_outputs = grub_term_outputs;
-
   FOR_ACTIVE_TERM_OUTPUTS (gfxterm)
     if (grub_strcmp (gfxterm->name, "gfxterm") == 0)
       break;
@@ -784,10 +784,17 @@ grub_test_use_gfxterm (void)
       return 1;
     }
 
+  if (gfxterm->init (gfxterm))
+    { 
+      grub_test_assert (0, "terminal `%s' failed: %s", "gfxterm", grub_errmsg);
+      return 1;
+    }
+
+  saved_outputs = grub_term_outputs;
   saved_gfxnext = gfxterm->next;
   grub_term_outputs = gfxterm;
   gfxterm->next = 0;
-  gfxterm->init (gfxterm);
+  use_gfxterm = 1;
 
   return 0;
 }
@@ -795,7 +802,12 @@ grub_test_use_gfxterm (void)
 void
 grub_test_use_gfxterm_end (void)
 {
+  if (!use_gfxterm)
+    return;
+  use_gfxterm = 0;
   gfxterm->fini (gfxterm);
   gfxterm->next = saved_gfxnext;
   grub_term_outputs = saved_outputs;
+  saved_outputs = 0;
+  saved_gfxnext = 0;
 }
