@@ -779,6 +779,10 @@ static struct grub_video_adapter grub_video_sm712_adapter =
     .get_palette = grub_video_fb_get_palette,
     .set_viewport = grub_video_fb_set_viewport,
     .get_viewport = grub_video_fb_get_viewport,
+    .set_region = grub_video_fb_set_region,
+    .get_region = grub_video_fb_get_region,
+    .set_area_status = grub_video_fb_set_area_status,
+    .get_area_status = grub_video_fb_get_area_status,
     .map_color = grub_video_fb_map_color,
     .map_rgb = grub_video_fb_map_rgb,
     .map_rgba = grub_video_fb_map_rgba,
@@ -796,20 +800,12 @@ static struct grub_video_adapter grub_video_sm712_adapter =
     .next = 0
   };
 
-#ifdef GRUB_MACHINE_MIPS_LOONGSON
-void grub_video_sm712_init (void)
-#else
 GRUB_MOD_INIT(video_sm712)
-#endif
 {
   grub_video_register (&grub_video_sm712_adapter);
 }
 
-#ifdef GRUB_MACHINE_MIPS_LOONGSON
-void grub_video_sm712_fini (void)
-#else
 GRUB_MOD_FINI(video_sm712)
-#endif
 {
   grub_video_unregister (&grub_video_sm712_adapter);
 }
