@@ -27,6 +27,7 @@
 
 #define	SCRATCH_PAD_BOOT		0x5000
 #define	SCRATCH_PAD_DISKBOOT		0x4000
+#define	SCRATCH_PAD_BOOT_SIZE		0x110
 
 #define GET_ABS(symbol, reg)	\
 	add	PIC_REG, (symbol - pic_base), reg
@@ -45,8 +46,7 @@
 
 #define GRUB_BOOT_MACHINE_KERNEL_BYTE 0x80
 
-#define GRUB_BOOT_MACHINE_CODE_END \
-	(0x1fc - GRUB_BOOT_AOUT_HEADER_SIZE)
+#define GRUB_BOOT_MACHINE_CODE_END 0x1fc
 
 #define GRUB_BOOT_MACHINE_KERNEL_ADDR 0x4200
 
@@ -56,7 +56,7 @@ struct grub_boot_blocklist
 {
   grub_uint64_t start;
   grub_uint32_t len;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 #endif
 
 #endif /* ! BOOT_MACHINE_HEADER */
