@@ -27,6 +27,8 @@
 #include <grub/fs.h>
 #include <grub/partition.h>
 #include <grub/msdos_partition.h>
+#include <grub/gpt_partition.h>
+#include <grub/i386/pc/boot.h>
 #include <grub/emu/hostdisk.h>
 #include <grub/emu/getroot.h>
 #include <grub/term.h>
@@ -42,11 +44,15 @@
 #include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
-#include <sys/stat.h>
 #include <assert.h>
 
 #define _GNU_SOURCE	1
+
+#pragma GCC diagnostic ignored "-Wmissing-prototypes"
+#pragma GCC diagnostic ignored "-Wmissing-declarations"
 #include <argp.h>
+#pragma GCC diagnostic error "-Wmissing-prototypes"
+#pragma GCC diagnostic error "-Wmissing-declarations"
 
 #include "progname.h"
 
@@ -57,6 +63,7 @@ enum {
   PRINT_DRIVE,
   PRINT_DEVICE,
   PRINT_PARTMAP,
+  PRINT_PARTUUID,
   PRINT_ABSTRACTION,
   PRINT_CRYPTODISK_UUID,
   PRINT_HINT_STR,
@@ -67,15 +74,87 @@ enum {
   PRINT_ARC_HINT,
   PRINT_COMPATIBILITY_HINT,
   PRINT_MSDOS_PARTTYPE,
+  PRINT_GPT_PARTTYPE,
   PRINT_ZERO_CHECK,
   PRINT_DISK
 };
 
+static const char *targets[] =
+  {
+    [PRINT_FS]                 = "fs",
+    [PRINT_FS_UUID]            = "fs_uuid",
+    [PRINT_FS_LABEL]           = "fs_label",
+    [PRINT_DRIVE]              = "drive",
+    [PRINT_DEVICE]             = "device",
+    [PRINT_PARTMAP]            = "partmap",
+    [PRINT_PARTUUID]           = "partuuid",
+    [PRINT_ABSTRACTION]        = "abstraction",
+    [PRINT_CRYPTODISK_UUID]    = "cryptodisk_uuid",
+    [PRINT_HINT_STR]           = "hints_string",
+    [PRINT_BIOS_HINT]          = "bios_hints",
+    [PRINT_IEEE1275_HINT]      = "ieee1275_hints",
+    [PRINT_BAREMETAL_HINT]     = "baremetal_hints",
+    [PRINT_EFI_HINT]           = "efi_hints",
+    [PRINT_ARC_HINT]           = "arc_hints",
+    [PRINT_COMPATIBILITY_HINT] = "compatibility_hint",
+    [PRINT_MSDOS_PARTTYPE]     = "msdos_parttype",
+    [PRINT_GPT_PARTTYPE]       = "gpt_parttype",
+    [PRINT_ZERO_CHECK]         = "zero_check",
+    [PRINT_DISK]               = "disk",
+  };
+
 static int print = PRINT_FS;
 static unsigned int argument_is_device = 0;
 
+static char *
+get_targets_string (void)
+{
+  char **arr = xmalloc (sizeof (targets));
+  int len = 0;
+  char *str;
+  char *ptr;
+  unsigned i;
+
+  memcpy (arr, targets, sizeof (targets));
+  qsort (arr, ARRAY_SIZE (targets), sizeof (char *), grub_qsort_strcmp);
+  for (i = 0; i < ARRAY_SIZE (targets); i++)
+    len += grub_strlen (targets[i]) + 2;
+  ptr = str = xmalloc (len);
+  for (i = 0; i < ARRAY_SIZE (targets); i++)
+    {
+      ptr = grub_stpcpy (ptr, arr[i]);
+      *ptr++ = ',';
+      *ptr++ = ' ';
+    }
+  ptr[-2] = '\0';
+  free (arr);
+
+  return str;
+}
+
+static int
+print_gpt_guid (grub_gpt_part_guid_t guid)
+{
+  guid.data1 = grub_le_to_cpu32 (guid.data1);
+  guid.data2 = grub_le_to_cpu16 (guid.data2);
+  guid.data3 = grub_le_to_cpu16 (guid.data3);
+
+  return grub_printf ("%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
+		      guid.data1, guid.data2, guid.data3, guid.data4[0],
+		      guid.data4[1], guid.data4[2], guid.data4[3],
+		      guid.data4[4], guid.data4[5], guid.data4[6],
+		      guid.data4[7]);
+}
+
 static void
-probe_partmap (grub_disk_t disk)
+do_print (const char *x, void *data)
+{
+  char delim = *(const char *) data;
+  grub_printf ("%s%c", x, delim);
+}
+
+static void
+probe_partmap (grub_disk_t disk, char delim)
 {
   grub_partition_t part;
   grub_disk_memberlist_t list = NULL, tmp;
@@ -86,19 +165,19 @@ probe_partmap (grub_disk_t disk)
     }
 
   for (part = disk->partition; part; part = part->parent)
-    printf ("%s ", part->partmap->name);
+    printf ("%s%c", part->partmap->name, delim);
 
   if (disk->dev->id == GRUB_DISK_DEVICE_DISKFILTER_ID)
-    grub_diskfilter_print_partmap (disk);
+    grub_diskfilter_get_partmap (disk, do_print, &delim);
 
   /* In case of LVM/RAID, check the member devices as well.  */
-  if (disk->dev->memberlist)
+  if (disk->dev->disk_memberlist)
     {
-      list = disk->dev->memberlist (disk);
+      list = disk->dev->disk_memberlist (disk);
     }
   while (list)
     {
-      probe_partmap (list->disk);
+      probe_partmap (list->disk, delim);
       tmp = list->next;
       free (list);
       list = tmp;
@@ -106,24 +185,66 @@ probe_partmap (grub_disk_t disk)
 }
 
 static void
-probe_cryptodisk_uuid (grub_disk_t disk)
+probe_partuuid (grub_disk_t disk, char delim)
+{
+  grub_partition_t p = disk->partition;
+
+  /*
+   * Nested partitions not supported for now.
+   * Non-nested partitions must have disk->partition->parent == NULL
+   */
+  if (p && p->parent == NULL)
+    {
+      disk->partition = p->parent;
+
+      if (strcmp(p->partmap->name, "msdos") == 0)
+	{
+	    /*
+	     * The partition GUID for MSDOS is the partition number (starting
+	     * with 1) prepended with the NT disk signature.
+	     */
+	    grub_uint32_t nt_disk_sig;
+
+	    if (grub_disk_read (disk, 0, GRUB_BOOT_MACHINE_WINDOWS_NT_MAGIC,
+				sizeof(nt_disk_sig), &nt_disk_sig) == 0)
+	      grub_printf ("%08x-%02x",
+			   grub_le_to_cpu32(nt_disk_sig), 1 + p->number);
+	}
+      else if (strcmp(p->partmap->name, "gpt") == 0)
+	{
+	  struct grub_gpt_partentry gptdata;
+
+	  if (grub_disk_read (disk, p->offset, p->index,
+			      sizeof(gptdata), &gptdata) == 0)
+	    print_gpt_guid(gptdata.guid);
+	}
+
+      disk->partition = p;
+    }
+}
+
+static void
+probe_cryptodisk_uuid (grub_disk_t disk, char delim)
 {
   grub_disk_memberlist_t list = NULL, tmp;
 
   /* In case of LVM/RAID, check the member devices as well.  */
-  if (disk->dev->memberlist)
+  if (disk->dev->disk_memberlist)
     {
-      list = disk->dev->memberlist (disk);
+      list = disk->dev->disk_memberlist (disk);
     }
   while (list)
     {
-      probe_cryptodisk_uuid (list->disk);
+      probe_cryptodisk_uuid (list->disk, delim);
       tmp = list->next;
       free (list);
       list = tmp;
     }
   if (disk->dev->id == GRUB_DISK_DEVICE_CRYPTODISK_ID)
-    grub_util_cryptodisk_print_uuid (disk);
+    {
+      const char *uu = grub_util_cryptodisk_get_uuid (disk);
+      grub_printf ("%s%c", uu, delim);
+    }
 }
 
 static int
@@ -145,146 +266,17 @@ probe_raid_level (grub_disk_t disk)
   return ((struct grub_diskfilter_lv *) disk->data)->segments->type;
 }
 
-/* Since OF path names can have "," characters in them, and GRUB
-   internally uses "," to indicate partitions (unlike OF which uses
-   ":" for this purpose) we escape such commas.  */
-static char *
-escape_of_path (const char *orig_path)
-{
-  char *new_path, *d, c;
-  const char *p;
-
-  if (!strchr (orig_path, ','))
-    return (char *) xstrdup (orig_path);
-
-  new_path = xmalloc (strlen (orig_path) * 2 + 1);
-
-  p = orig_path;
-  d = new_path;
-  while ((c = *p++) != '\0')
-    {
-      if (c == ',')
-	*d++ = '\\';
-      *d++ = c;
-    }
-  *d = 0;
-
-  return new_path;
-}
-
-static char *
-guess_bios_drive (const char *orig_path)
-{
-  char *canon;
-  char *ptr;
-  canon = canonicalize_file_name (orig_path);
-  if (!canon)
-    return NULL;
-  ptr = strrchr (orig_path, '/');
-  if (ptr)
-    ptr++;
-  else
-    ptr = canon;
-  if ((ptr[0] == 's' || ptr[0] == 'h') && ptr[1] == 'd')
-    {
-      int num = ptr[2] - 'a';
-      free (canon);
-      return xasprintf ("hd%d", num);
-    }
-  if (ptr[0] == 'f' && ptr[1] == 'd')
-    {
-      int num = atoi (ptr + 2);
-      free (canon);
-      return xasprintf ("fd%d", num);
-    }
-  free (canon);
-  return NULL;
-}
-
-static char *
-guess_efi_drive (const char *orig_path)
-{
-  char *canon;
-  char *ptr;
-  canon = canonicalize_file_name (orig_path);
-  if (!canon)
-    return NULL;
-  ptr = strrchr (orig_path, '/');
-  if (ptr)
-    ptr++;
-  else
-    ptr = canon;
-  if ((ptr[0] == 's' || ptr[0] == 'h') && ptr[1] == 'd')
-    {
-      int num = ptr[2] - 'a';
-      free (canon);
-      return xasprintf ("hd%d", num);
-    }
-  if (ptr[0] == 'f' && ptr[1] == 'd')
-    {
-      int num = atoi (ptr + 2);
-      free (canon);
-      return xasprintf ("fd%d", num);
-    }
-  free (canon);
-  return NULL;
-}
-
-static char *
-guess_baremetal_drive (const char *orig_path)
-{
-  char *canon;
-  char *ptr;
-  canon = canonicalize_file_name (orig_path);
-  if (!canon)
-    return NULL;
-  ptr = strrchr (orig_path, '/');
-  if (ptr)
-    ptr++;
-  else
-    ptr = canon;
-  if (ptr[0] == 'h' && ptr[1] == 'd')
-    {
-      int num = ptr[2] - 'a';
-      free (canon);
-      return xasprintf ("ata%d", num);
-    }
-  if (ptr[0] == 's' && ptr[1] == 'd')
-    {
-      int num = ptr[2] - 'a';
-      free (canon);
-      return xasprintf ("ahci%d", num);
-    }
-  free (canon);
-  return NULL;
-}
-
 static void
-print_full_name (const char *drive, grub_device_t dev)
-{
-  char *dname = escape_of_path (drive);
-  if (dev->disk->partition)
-    {
-      char *pname = grub_partition_get_name (dev->disk->partition);
-      printf ("%s,%s", dname, pname);
-      free (pname);
-    }
-  else
-    printf ("%s", dname);
-  free (dname);
-} 
-
-static void
-probe_abstraction (grub_disk_t disk)
+probe_abstraction (grub_disk_t disk, char delim)
 {
   grub_disk_memberlist_t list = NULL, tmp;
   int raid_level;
 
-  if (disk->dev->memberlist)
-    list = disk->dev->memberlist (disk);
+  if (disk->dev->disk_memberlist)
+    list = disk->dev->disk_memberlist (disk);
   while (list)
     {
-      probe_abstraction (list->disk);
+      probe_abstraction (list->disk, delim);
 
       tmp = list->next;
       free (list);
@@ -292,27 +284,28 @@ probe_abstraction (grub_disk_t disk)
     }
 
   if (disk->dev->id == GRUB_DISK_DEVICE_DISKFILTER_ID
-      && grub_memcmp (disk->name, "lvm/", sizeof ("lvm/") - 1) == 0)
-    printf ("lvm ");
+      && (grub_memcmp (disk->name, "lvm/", sizeof ("lvm/") - 1) == 0 ||
+	  grub_memcmp (disk->name, "lvmid/", sizeof ("lvmid/") - 1) == 0))
+    printf ("lvm%c", delim);
 
   if (disk->dev->id == GRUB_DISK_DEVICE_DISKFILTER_ID
       && grub_memcmp (disk->name, "ldm/", sizeof ("ldm/") - 1) == 0)
-    printf ("ldm ");
+    printf ("ldm%c", delim);
 
   if (disk->dev->id == GRUB_DISK_DEVICE_CRYPTODISK_ID)
-    grub_util_cryptodisk_print_abstraction (disk);
+    grub_util_cryptodisk_get_abstraction (disk, do_print, &delim);
 
   raid_level = probe_raid_level (disk);
   if (raid_level >= 0)
     {
-      printf ("diskfilter ");
-      if (disk->dev->raidname)
-	printf ("%s ", disk->dev->raidname (disk));
+      printf ("diskfilter%c", delim);
+      if (disk->dev->disk_raidname)
+	printf ("%s%c", disk->dev->disk_raidname (disk), delim);
     }
   if (raid_level == 5)
-    printf ("raid5rec ");
+    printf ("raid5rec%c", delim);
   if (raid_level == 6)
-    printf ("raid6rec ");
+    printf ("raid6rec%c", delim);
 }
 
 static void
@@ -325,9 +318,9 @@ probe (const char *path, char **device_names, char delim)
 
   if (path != NULL)
     {
-      grub_path = canonicalize_file_name (path);
+      grub_path = grub_canonicalize_file_name (path);
       if (! grub_path)
-	grub_util_error (_("failed to get canonical path of %s"), path);
+	grub_util_error (_("failed to get canonical path of `%s'"), path);
       device_names = grub_guess_root_devices (grub_path);
       free (grub_path);
     }
@@ -342,7 +335,7 @@ probe (const char *path, char **device_names, char delim)
 	  printf ("%s", *curdev);
 	  putchar (delim);
 	}
-      return;
+      goto free_device_names;
     }
 
   if (print == PRINT_DISK)
@@ -358,8 +351,9 @@ probe (const char *path, char **device_names, char delim)
 	    }
 	  printf ("%s", disk);
 	  putchar (delim);
+	  free (disk);
 	}
-      return;
+      goto free_device_names;
     }
 
   for (curdev = device_names; *curdev; curdev++)
@@ -367,8 +361,8 @@ probe (const char *path, char **device_names, char delim)
       grub_util_pull_device (*curdev);
       ndev++;
     }
-  
-  drives_names = xmalloc (sizeof (drives_names[0]) * (ndev + 1)); 
+
+  drives_names = xcalloc (ndev + 1, sizeof (drives_names[0]));
 
   for (curdev = device_names, curdrive = drives_names; *curdev; curdev++,
        curdrive++)
@@ -404,7 +398,7 @@ probe (const char *path, char **device_names, char delim)
 	  if (! dev || !dev->disk)
 	    grub_util_error ("%s", grub_errmsg);
 
-	  dsize = grub_disk_get_size (dev->disk);
+	  dsize = grub_disk_native_sectors (dev->disk);
 	  for (addr = 0; addr < dsize;
 	       addr += sizeof (buffer) / GRUB_DISK_SECTOR_SIZE)
 	    {
@@ -452,10 +446,10 @@ probe (const char *path, char **device_names, char delim)
       else if (print == PRINT_FS_UUID)
 	{
 	  char *uuid;
-	  if (! fs->uuid)
+	  if (! fs->fs_uuid)
 	    grub_util_error (_("%s does not support UUIDs"), fs->name);
 
-	  if (fs->uuid (dev, &uuid) != GRUB_ERR_NONE)
+	  if (fs->fs_uuid (dev, &uuid) != GRUB_ERR_NONE)
 	    grub_util_error ("%s", grub_errmsg);
 
 	  printf ("%s", uuid);
@@ -464,16 +458,17 @@ probe (const char *path, char **device_names, char delim)
       else if (print == PRINT_FS_LABEL)
 	{
 	  char *label;
-	  if (! fs->label)
+	  if (! fs->fs_label)
 	    grub_util_error (_("filesystem `%s' does not support labels"),
 			     fs->name);
 
-	  if (fs->label (dev, &label) != GRUB_ERR_NONE)
+	  if (fs->fs_label (dev, &label) != GRUB_ERR_NONE)
 	    grub_util_error ("%s", grub_errmsg);
 
 	  printf ("%s", label);
 	  putchar (delim);
 	}
+      grub_device_close (dev);
       goto end;
     }
 
@@ -490,7 +485,7 @@ probe (const char *path, char **device_names, char delim)
       if (print == PRINT_HINT_STR)
 	{
 	  const char *osdev = grub_util_biosdisk_get_osdev (dev->disk);
-	  const char *ofpath = osdev ? grub_util_devname_to_ofpath (osdev) : 0;
+	  char *ofpath = osdev ? grub_util_devname_to_ofpath (osdev) : 0;
 	  char *biosname, *bare, *efi;
 	  const char *map;
 
@@ -498,37 +493,38 @@ probe (const char *path, char **device_names, char delim)
 	    {
 	      char *tmp = xmalloc (strlen (ofpath) + sizeof ("ieee1275/"));
 	      char *p;
-	      p = stpcpy (tmp, "ieee1275/");
+	      p = grub_stpcpy (tmp, "ieee1275/");
 	      strcpy (p, ofpath);
 	      printf ("--hint-ieee1275='");
-	      print_full_name (tmp, dev);
+	      grub_util_fprint_full_disk_name (stdout, tmp, dev);
 	      printf ("' ");
 	      free (tmp);
+	      free (ofpath);
 	    }
 
-	  biosname = guess_bios_drive (*curdev);
+	  biosname = grub_util_guess_bios_drive (*curdev);
 	  if (biosname)
 	    {
 	      printf ("--hint-bios=");
-	      print_full_name (biosname, dev);
+	      grub_util_fprint_full_disk_name (stdout, biosname, dev);
 	      printf (" ");
 	    }
 	  free (biosname);
 
-	  efi = guess_efi_drive (*curdev);
+	  efi = grub_util_guess_efi_drive (*curdev);
 	  if (efi)
 	    {
 	      printf ("--hint-efi=");
-	      print_full_name (efi, dev);
+	      grub_util_fprint_full_disk_name (stdout, efi, dev);
 	      printf (" ");
 	    }
 	  free (efi);
 
-	  bare = guess_baremetal_drive (*curdev);
+	  bare = grub_util_guess_baremetal_drive (*curdev);
 	  if (bare)
 	    {
 	      printf ("--hint-baremetal=");
-	      print_full_name (bare, dev);
+	      grub_util_fprint_full_disk_name (stdout, bare, dev);
 	      printf (" ");
 	    }
 	  free (bare);
@@ -539,75 +535,70 @@ probe (const char *path, char **device_names, char delim)
 	  if (map)
 	    {
 	      printf ("--hint='");
-	      print_full_name (map, dev);
+	      grub_util_fprint_full_disk_name (stdout, map, dev);
 	      printf ("' ");
 	    }
-	  printf ("\n");
-
-	  grub_device_close (dev);
-	  continue;
+	  if (curdrive[1])
+	    printf (" ");
+	  else
+	    printf ("\n");
 	}
       
-      if ((print == PRINT_COMPATIBILITY_HINT || print == PRINT_BIOS_HINT
+      else if ((print == PRINT_COMPATIBILITY_HINT || print == PRINT_BIOS_HINT
 	   || print == PRINT_IEEE1275_HINT || print == PRINT_BAREMETAL_HINT
 	   || print == PRINT_EFI_HINT || print == PRINT_ARC_HINT)
 	  && dev->disk->dev->id != GRUB_DISK_DEVICE_HOSTDISK_ID)
 	{
-	  print_full_name (dev->disk->name, dev);
+	  grub_util_fprint_full_disk_name (stdout, dev->disk->name, dev);
 	  putchar (delim);
-	  continue;
 	}
 
-      if (print == PRINT_COMPATIBILITY_HINT)
+      else if (print == PRINT_COMPATIBILITY_HINT)
 	{
 	  const char *map;
 	  char *biosname;
 	  map = grub_util_biosdisk_get_compatibility_hint (dev->disk);
 	  if (map)
 	    {
-	      print_full_name (map, dev);
+	      grub_util_fprint_full_disk_name (stdout, map, dev);
 	      putchar (delim);
 	      grub_device_close (dev);
 	      /* Compatibility hint is one device only.  */
 	      break;
 	    }
-	  biosname = guess_bios_drive (*curdev);
+	  biosname = grub_util_guess_bios_drive (*curdev);
 	  if (biosname)
 	    {
-	      print_full_name (biosname, dev);
+	      grub_util_fprint_full_disk_name (stdout, biosname, dev);
 	      putchar (delim);
+	      free (biosname);
+	      /* Compatibility hint is one device only.  */
+	      grub_device_close (dev);
+	      break;
 	    }
-	  free (biosname);
-	  grub_device_close (dev);
-	  /* Compatibility hint is one device only.  */
-	  if (biosname)
-	    break;
-	  continue;
 	}
 
-      if (print == PRINT_BIOS_HINT)
+      else if (print == PRINT_BIOS_HINT)
 	{
 	  char *biosname;
-	  biosname = guess_bios_drive (*curdev);
+	  biosname = grub_util_guess_bios_drive (*curdev);
 	  if (biosname)
 	    {
-	      print_full_name (biosname, dev);
+	      grub_util_fprint_full_disk_name (stdout, biosname, dev);
 	      putchar (delim);
+	      free (biosname);
 	    }
-	  free (biosname);
-	  grub_device_close (dev);
-	  continue;
 	}
-      if (print == PRINT_IEEE1275_HINT)
+      else if (print == PRINT_IEEE1275_HINT)
 	{
 	  const char *osdev = grub_util_biosdisk_get_osdev (dev->disk);
-	  const char *ofpath = grub_util_devname_to_ofpath (osdev);
+	  char *ofpath = grub_util_devname_to_ofpath (osdev);
 	  const char *map;
 
 	  map = grub_util_biosdisk_get_compatibility_hint (dev->disk);
 	  if (map)
 	    {
-	      print_full_name (map, dev);
+	      grub_util_fprint_full_disk_name (stdout, map, dev);
 	      putchar (delim);
 	    }
 
@@ -615,120 +606,124 @@ probe (const char *path, char **device_names, char delim)
 	    {
 	      char *tmp = xmalloc (strlen (ofpath) + sizeof ("ieee1275/"));
 	      char *p;
-	      p = stpcpy (tmp, "ieee1275/");
+	      p = grub_stpcpy (tmp, "ieee1275/");
 	      strcpy (p, ofpath);
-	      print_full_name (tmp, dev);
+	      grub_util_fprint_full_disk_name (stdout, tmp, dev);
 	      free (tmp);
+	      free (ofpath);
 	      putchar (delim);
 	    }
-
-	  grub_device_close (dev);
-	  continue;
 	}
-      if (print == PRINT_EFI_HINT)
+      else if (print == PRINT_EFI_HINT)
 	{
 	  char *biosname;
 	  const char *map;
-	  biosname = guess_efi_drive (*curdev);
+	  biosname = grub_util_guess_efi_drive (*curdev);
 
 	  map = grub_util_biosdisk_get_compatibility_hint (dev->disk);
 	  if (map)
 	    {
-	      print_full_name (map, dev);
+	      grub_util_fprint_full_disk_name (stdout, map, dev);
 	      putchar (delim);
 	    }
 	  if (biosname)
 	    {
-	      print_full_name (biosname, dev);
+	      grub_util_fprint_full_disk_name (stdout, biosname, dev);
 	      putchar (delim);
+	      free (biosname);
 	    }
-
-	  free (biosname);
-	  grub_device_close (dev);
-	  continue;
 	}
 
-      if (print == PRINT_BAREMETAL_HINT)
+      else if (print == PRINT_BAREMETAL_HINT)
 	{
 	  char *biosname;
 	  const char *map;
 
-	  biosname = guess_baremetal_drive (*curdev);
+	  biosname = grub_util_guess_baremetal_drive (*curdev);
 
 	  map = grub_util_biosdisk_get_compatibility_hint (dev->disk);
 	  if (map)
 	    {
-	      print_full_name (map, dev);
+	      grub_util_fprint_full_disk_name (stdout, map, dev);
 	      putchar (delim);
 	    }
 	  if (biosname)
 	    {
-	      print_full_name (biosname, dev);
+	      grub_util_fprint_full_disk_name (stdout, biosname, dev);
 	      putchar (delim);
+	      free (biosname);
 	    }
-
-	  free (biosname);
-	  grub_device_close (dev);
-	  continue;
 	}
 
-      if (print == PRINT_ARC_HINT)
+      else if (print == PRINT_ARC_HINT)
 	{
 	  const char *map;
 
 	  map = grub_util_biosdisk_get_compatibility_hint (dev->disk);
 	  if (map)
 	    {
-	      print_full_name (map, dev);
+	      grub_util_fprint_full_disk_name (stdout, map, dev);
 	      putchar (delim);
 	    }
-
-	  /* FIXME */
-	  grub_device_close (dev);
-	  continue;
 	}
 
-      if (print == PRINT_ABSTRACTION)
+      else if (print == PRINT_ABSTRACTION)
+	probe_abstraction (dev->disk, delim);
+
+      else if (print == PRINT_CRYPTODISK_UUID)
+	probe_cryptodisk_uuid (dev->disk, delim);
+
+      else if (print == PRINT_PARTMAP)
+	/* Check if dev->disk itself is contained in a partmap.  */
+	probe_partmap (dev->disk, delim);
+
+      else if (print == PRINT_PARTUUID)
 	{
-	  probe_abstraction (dev->disk);
+	  probe_partuuid (dev->disk, delim);
 	  putchar (delim);
-	  grub_device_close (dev);
-	  continue;
 	}
 
-      if (print == PRINT_CRYPTODISK_UUID)
-	{
-	  probe_cryptodisk_uuid (dev->disk);
-	  putchar (delim);
-	  grub_device_close (dev);
-	  continue;
-	}
-
-      if (print == PRINT_PARTMAP)
-	{
-	  /* Check if dev->disk itself is contained in a partmap.  */
-	  probe_partmap (dev->disk);
-	  putchar (delim);
-	  grub_device_close (dev);
-	  continue;
-	}
-
-      if (print == PRINT_MSDOS_PARTTYPE)
+      else if (print == PRINT_MSDOS_PARTTYPE)
 	{
 	  if (dev->disk->partition
 	      && strcmp(dev->disk->partition->partmap->name, "msdos") == 0)
 	    printf ("%02x", dev->disk->partition->msdostype);
 
 	  putchar (delim);
-	  grub_device_close (dev);
-	  continue;
 	}
+
+      else if (print == PRINT_GPT_PARTTYPE)
+	{
+          if (dev->disk->partition
+	      && strcmp (dev->disk->partition->partmap->name, "gpt") == 0)
+            {
+              struct grub_gpt_partentry gptdata;
+              grub_partition_t p = dev->disk->partition;
+              dev->disk->partition = dev->disk->partition->parent;
+
+              if (grub_disk_read (dev->disk, p->offset, p->index,
+                                  sizeof (gptdata), &gptdata) == 0)
+		print_gpt_guid(gptdata.type);
+              dev->disk->partition = p;
+            }
+          putchar (delim);
+        }
+
+      grub_device_close (dev);
     }
 
  end:
   for (curdrive = drives_names; *curdrive; curdrive++)
     free (*curdrive);
   free (drives_names);
+
+free_device_names:
+  if (path != NULL)
+    {
+      for (curdev = device_names; *curdev; curdev++)
+	free (*curdev);
+      free (device_names);
+    }
 }
 
 static struct argp_option options[] = {
@@ -736,12 +731,14 @@ static struct argp_option options[] = {
    N_("given argument is a system device, not a path"), 0},
   {"device-map",  'm', N_("FILE"), 0,
    N_("use FILE as the device map [default=%s]"), 0},
-  {"target",  't', "(fs|fs_uuid|fs_label|drive|device|partmap|abstraction|cryptodisk_uuid|msdos_parttype)", 0,
-   N_("print filesystem module, GRUB drive, system device, partition map module, abstraction module or cryptographic container UUID [default=fs]"), 0},
+  {"target",  't', N_("TARGET"), 0, 0, 0},
   {"verbose",     'v', 0,      0, N_("print verbose messages."), 0},
+  {0, '0', 0, 0, N_("separate items in output using ASCII NUL characters"), 0},
   { 0, 0, 0, 0, 0, 0 }
 };
 
+#pragma GCC diagnostic ignored "-Wformat-nonliteral"
+
 static char *
 help_filter (int key, const char *text, void *input __attribute__ ((unused)))
 {
@@ -750,11 +747,26 @@ help_filter (int key, const char *text, void *input __attribute__ ((unused)))
       case 'm':
         return xasprintf (text, DEFAULT_DEVICE_MAP);
 
+      case 't':
+	{
+	  char *ret, *t = get_targets_string (), *def;
+
+	  def = xasprintf (_("[default=%s]"), targets[print]);
+
+	  ret = xasprintf ("%s\n%s %s %s", _("print TARGET"),
+			    _("available targets:"), t, def);
+	  free (t);
+	  free (def);
+	  return ret;
+	}
+
       default:
         return (char *) text;
     }
 }
 
+#pragma GCC diagnostic error "-Wformat-nonliteral"
+
 struct arguments
 {
   char **devices;
@@ -785,44 +797,18 @@ argp_parser (int key, char *arg, struct argp_state *state)
       break;
 
     case 't':
-      if (!strcmp (arg, "fs"))
-	print = PRINT_FS;
-      else if (!strcmp (arg, "fs_uuid"))
-	print = PRINT_FS_UUID;
-      else if (!strcmp (arg, "fs_label"))
-	print = PRINT_FS_LABEL;
-      else if (!strcmp (arg, "drive"))
-	print = PRINT_DRIVE;
-      else if (!strcmp (arg, "device"))
-	print = PRINT_DEVICE;
-      else if (!strcmp (arg, "partmap"))
-	print = PRINT_PARTMAP;
-      else if (!strcmp (arg, "abstraction"))
-	print = PRINT_ABSTRACTION;
-      else if (!strcmp (arg, "cryptodisk_uuid"))
-	print = PRINT_CRYPTODISK_UUID;
-      else if (!strcmp (arg, "msdos_parttype"))
-	print = PRINT_MSDOS_PARTTYPE;
-      else if (!strcmp (arg, "hints_string"))
-	print = PRINT_HINT_STR;
-      else if (!strcmp (arg, "bios_hints"))
-	print = PRINT_BIOS_HINT;
-      else if (!strcmp (arg, "ieee1275_hints"))
-	print = PRINT_IEEE1275_HINT;
-      else if (!strcmp (arg, "baremetal_hints"))
-	print = PRINT_BAREMETAL_HINT;
-      else if (!strcmp (arg, "efi_hints"))
-	print = PRINT_EFI_HINT;
-      else if (!strcmp (arg, "arc_hints"))
-	print = PRINT_ARC_HINT;
-      else if (!strcmp (arg, "compatibility_hint"))
-	print = PRINT_COMPATIBILITY_HINT;
-      else if (strcmp (arg, "zero_check") == 0)
-	print = PRINT_ZERO_CHECK;
-      else if (!strcmp (arg, "disk"))
-	print = PRINT_DISK;
-      else
-	argp_usage (state);
+      {
+	int i;
+
+	for (i = PRINT_FS; i < ARRAY_SIZE (targets); i++)
+	  if (strcmp (arg, targets[i]) == 0)
+	    {
+	      print = i;
+	      break;
+	    }
+	if (i == ARRAY_SIZE (targets))
+	  argp_usage (state);
+      }
       break;
 
     case '0':
@@ -862,9 +848,7 @@ main (int argc, char *argv[])
   char delim;
   struct arguments arguments;
 
-  set_program_name (argv[0]);
-
-  grub_util_init_nls ();
+  grub_util_host_init (&argc, &argv);
 
   memset (&arguments, 0, sizeof (struct arguments));
   arguments.device_max = argc + 1;
@@ -926,11 +910,7 @@ main (int argc, char *argv[])
   else
     probe (arguments.devices[0], NULL, delim);
 
-  if (!arguments.zero_delim && (print == PRINT_BIOS_HINT
-				|| print == PRINT_IEEE1275_HINT
-				|| print == PRINT_BAREMETAL_HINT
-				|| print == PRINT_EFI_HINT
-				|| print == PRINT_ARC_HINT))
+  if (delim == ' ')
     putchar ('\n');
 
   /* Free resources.  */
